/* direct.cpp - Class for handling a directory listing
 * 
 *  Bronco File Manager
 * 
 *  Copyright (c) 2000, 2001 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include "direct.h"

Directory::Directory(){
   top = cur  = NULL;

   sort_order = (SORT_ASCII | DIRS_FIRST | SHOW_DOTS);
   m_error    = NULL;
   m_errno    = 0;

   return;
}

Directory::~Directory(){

   clear();
   
   return;
}

int Directory::clear(void){
   struct direct *temp;

   cur = top;
   while(cur){
      temp = cur->next;
      free(cur->name);
      delete cur;
      cur = temp;
   }

   cur = top = NULL;
   return(0);
}

int Directory::add(const char *n, struct stat *statbuf){
   struct direct *temp;
   temp = new struct direct;

   temp->name = strdup(n);
   memcpy(&(temp->dirdata), statbuf, sizeof(struct stat));
   temp->next = NULL;
   temp->prev = NULL;

   struct direct *mcur, *mlast;

   mlast = NULL;
   mcur = top;
   while(mcur && compare(temp, mcur) > 0){
      mlast = mcur;
      mcur = mcur->next;
   }

   if(mlast){
      if(mcur){
         temp->next  = mcur;
         temp->prev  = mlast;
         mlast->next = temp;
         mcur->prev  = temp;
      }
      else{
         temp->prev  = mlast;
         mlast->next = temp;
      }
   }
   else{
      if(mcur){
         temp->next  = mcur;
         mcur->prev  = temp;
      }
      else{
      }
      top = temp;
   }

   cur = temp;

   return(1);
}

int Directory::add(const char *n){
   struct stat statbuf;
   if(lstat(n, &statbuf) >= 0){
      add(n, &statbuf);
      return(1);
   }
   else{
      return(0);
   }
}

int Directory::find(const char *n, struct stat *d){
   cur = top;
   while(cur){
      if(strcmp(cur->name, n) == 0){
         memcpy(d, &(cur->dirdata), sizeof(struct stat));
         return(1);
      }
      cur = cur->next;
   }
   cur = top;
   return(0);
}

int Directory::first(char *n, struct stat *d){
   cur = top;
   if(cur){
      strcpy(n, cur->name);
      memcpy(d, &(cur->dirdata), sizeof(struct stat));
      return(1);
   }
   return(0);
}

int Directory::next(char *n, struct stat *d){
   if(cur && cur->next){
      cur = cur->next;
      strcpy(n, cur->name);
      memcpy(d, &(cur->dirdata), sizeof(struct stat));
      return(1);
   }
   return(0);
}

int Directory::prev(char *n, struct stat *d){
   if(cur && cur->prev){
      cur = cur->prev;
      d = &(cur->dirdata);
      n = cur->name;
      return(1);
   }
   return(0);
}

int Directory::setSorting(int order){
   sort_order = order;
   sort();
   return(0);
}

int Directory::compare(struct direct *d1, struct direct *d2){
   int rval  = 0;
   int isdir = 0;

   if(sort_order & DIRS_FIRST){
      if((S_ISDIR(d1->dirdata.st_mode) || S_ISSDIR(d1->dirdata.st_mode)) && !(S_ISDIR(d2->dirdata.st_mode) || S_ISSDIR(d2->dirdata.st_mode))){
         rval  = -1;
         isdir =  1;
      }
      if(!(S_ISDIR(d1->dirdata.st_mode) || S_ISSDIR(d1->dirdata.st_mode)) && (S_ISDIR(d2->dirdata.st_mode) || S_ISSDIR(d2->dirdata.st_mode))){
         rval  = +1;
         isdir =  1;
      }
   }

   int mode1, mode2;
   int m1, m2;

   if(!rval){
      switch(sort_order & SORT_MASK){
         case SORT_ASCII:
            rval = strcmp(d1->name, d2->name);
            break;
         case SORT_ALPHA:
            rval = strcasecmp(d1->name, d2->name);
            break;
         case SORT_MTIME:
            if(d1->dirdata.st_mtime > d2->dirdata.st_mtime) rval = -1;
            if(d1->dirdata.st_mtime < d2->dirdata.st_mtime) rval = +1;
            break;
         case SORT_ATIME:
            if(d1->dirdata.st_atime > d2->dirdata.st_atime) rval = -1;
            if(d1->dirdata.st_atime < d2->dirdata.st_atime) rval = +1;
            break;
         case SORT_PERM:
            mode1 = d1->dirdata.st_mode;
            mode2 = d2->dirdata.st_mode;

            mode1 = (mode1 & (S_IRWXU | S_IRWXG | S_IRWXO));
            mode2 = (mode2 & (S_IRWXU | S_IRWXG | S_IRWXO));

            m1 = (((mode1&S_IRWXU) ? 4 : 0) | ((mode1&S_IRWXG) ? 2 : 0) | ((mode1&S_IRWXO) ? 1 : 0));
            m2 = (((mode2&S_IRWXU) ? 4 : 0) | ((mode2&S_IRWXG) ? 2 : 0) | ((mode2&S_IRWXO) ? 1 : 0));

            if(m1 < m2){
               rval = -1;
            }
            else if(m1 > m2){
               rval = +1;
            }
            else{
               if(mode1 < mode2) rval = -1;
               if(mode1 > mode2) rval = +1;
            }
            break;
         case SORT_SIZE:
            if(d1->dirdata.st_size < d2->dirdata.st_size) rval = -1;
            if(d1->dirdata.st_size > d2->dirdata.st_size) rval = +1;
            break;
      }
   }

   if(isdir && !(sort_order & DIR_REVERSE)) return(rval);
   else      return((sort_order & SORT_REVERSE) ? -rval : rval);
}

int Directory::print(){
   struct direct *mcur;
   mcur = top;
   while(mcur){
      printf("%c ", (S_ISDIR(mcur->dirdata.st_mode) ? 'd' : ' '));

      if(S_IRUSR & mcur->dirdata.st_mode) printf("r");
      else                                printf("-");
      if(S_IWUSR & mcur->dirdata.st_mode) printf("w");
      else                                printf("-");
      if(S_IXUSR & mcur->dirdata.st_mode) printf("x");
      else                                printf("-");

      if(S_IRGRP & mcur->dirdata.st_mode) printf("r");
      else                                printf("-");
      if(S_IWGRP & mcur->dirdata.st_mode) printf("w");
      else                                printf("-");
      if(S_IXGRP & mcur->dirdata.st_mode) printf("x");
      else                                printf("-");

      if(S_IROTH & mcur->dirdata.st_mode) printf("r");
      else                                printf("-");
      if(S_IWOTH & mcur->dirdata.st_mode) printf("w");
      else                                printf("-");
      if(S_IXOTH & mcur->dirdata.st_mode) printf("x");
      else                                printf("-");

      printf(" %s\n", mcur->name);
      mcur = mcur->next;
   }
   return(0);
}

int Directory::size(){
   int count=0;

   struct direct *mcur;
   for(mcur=top, count=0; mcur; mcur=mcur->next, count++)
      /* Empty loop */;
   
   return(count * sizeof(struct direct));
}

int Directory::loadDir(const char *dirname){
   DIR *dp;
   struct dirent *d;
   struct stat   s;

   char dirbuf[4096];

   if((dp = opendir(dirname)) != NULL){
      while((d = readdir(dp)) != NULL){
         if((!(sort_order & SHOW_DOTS) && !strncmp(".", d->d_name, 1))) 
            continue;
         if((!(sort_order & SHOW_IMPLIED) && (!strcmp(".", d->d_name) || !strcmp("..", d->d_name))) ) 
            continue;

         snprintf(dirbuf, sizeof(dirbuf), "%s/%s", dirname, d->d_name);
         if(lstat(dirbuf, &s) >= 0){
            if(S_ISLNK(s.st_mode) && isDir(dirbuf)){
               s.st_mode = s.st_mode | S_IFDIR;
            }

            if((sort_order & DIRS_ONLY) && !(S_ISDIR(s.st_mode) || S_ISSDIR(s.st_mode))) 
               continue;

            add(d->d_name, &s);
         }
         else
         {
            m_errno = errno;
            m_error = strerror(errno);
            closedir(dp);
            return(-1);
         }
      }
      closedir(dp);
   }
   else{
      m_errno = errno;
      m_error = strerror(errno);
      return(-1);
   }
   return(0);
}

int isDir(char *d){
   DIR *dp;
   if((dp = opendir(d)) != 0){
      closedir(dp);
      return(1);
   }
   else{
      return(0);
   }
}

int Directory::sort(void){
   struct direct *temp;
   struct direct *swap;
   struct direct *tempnext;

   int changed = 0;

   temp = top;
   while(temp){
      if(temp->next && (compare(temp, temp->next) > 0)){
         changed = 1;
         tempnext = temp->next;
         
         if(temp->prev)
            temp->prev->next = tempnext;
         else
            top = tempnext;

         if(tempnext->next)
            tempnext->next->prev = temp;

         swap = temp->next;
         temp->next = tempnext->next;
         tempnext->next = temp;
         
         swap = temp->prev;
         temp->prev = tempnext;
         tempnext->prev = swap;
      }
      else{
         temp = temp->next;
      }

      if(temp == NULL) {
         if(changed){
            temp = top;
            changed = 0;
         }
      }
   }
   return(0);
}

const char * Directory::errorString(void)
{
   return m_error;
}

int Directory::errorNumber(void)
{
   return m_errno;
}

