/* errlist.cpp - Generic window for editing lists
 * 
 *  Bronco File Manager
 * 
 *  Copyright (c) 2000, 2001 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include "errlist.h"

#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <qheader.h>
#include <qlistview.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <qvariant.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qstringlist.h>
#include <qmessagebox.h>

#include "errtext.h"
#include "prefs.h"
#include "bronco.h"

static const char* arrowdown_xpm[]={
    "17 16 4 1",
    "c c #98a6ff", "o c #000000", "@ c #000000", ". c None",
    "..................",
    "......oooooo......",
    "......oooooo......",
    "......oooooo......",
    "......oooooo......",
    "......oooooo......",
    "......oooooo......",
    "..oooooooooooooo..",
    "..oooooooooooooo..",
    "...oooooooooooo...",
    "....oooooooooo....",
    ".....oooooooo.....",
    "......oooooo......",
    ".......oooo.......",
    "........oo........",
    ".................."};

static const char* arrowup_xpm[]={
    "17 16 4 1",
    "c c #98a6ff", "o c #000000", "@ c #000000", ". c None",
    "..................",
    "........oo........",
    ".......oooo.......",
    "......oooooo......",
    ".....oooooooo.....",
    "....oooooooooo....",
    "...oooooooooooo...",
    "..oooooooooooooo..",
    "..oooooooooooooo..",
    "......oooooo......",
    "......oooooo......",
    "......oooooo......",
    "......oooooo......",
    "......oooooo......",
    "......oooooo......",
    ".................."};

ErrList::ErrList( QWidget* parent,  const char* name, WFlags fl )
    : QWidget( parent, name, fl )
{
   if ( !name )
      setName( "ErrList" );

   resize( 478, 274 ); 
   setCaption( tr( "List Edit" ) );

   ErrListLayout = new QVBoxLayout( this );
   ErrListLayout->setMargin(11);
   ErrListLayout->setSpacing(6);

   list_v = new QListView( this, "list_v" );
   list_v->setAllColumnsShowFocus( true );
   list_v->setSorting(-1, false);
   list_v->header()->setClickEnabled(false);
   list_v->header()->setMovingEnabled( false );
   list_v->addColumn("Log Name");
   ErrListLayout->addWidget( list_v );

   Layout2 = new QHBoxLayout( ErrListLayout ); 
   Layout2->setSpacing( 6 );
   Layout2->setMargin( 0 );

   up_b = new QPushButton( this, "up_b" );
   up_b->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)1, (QSizePolicy::SizeType)1, up_b->sizePolicy().hasHeightForWidth() ) );
   up_b->setPixmap( arrowup_xpm );
   up_b->hide();
   Layout2->addWidget( up_b );

   down_b = new QPushButton( this, "down_b" );
   down_b->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)1, (QSizePolicy::SizeType)1, down_b->sizePolicy().hasHeightForWidth() ) );
   down_b->setPixmap( arrowdown_xpm );
   down_b->hide();
   Layout2->addWidget( down_b );

   QSpacerItem* spacer = new QSpacerItem( 20, 20, QSizePolicy::Expanding, QSizePolicy::Minimum );
   Layout2->addItem( spacer );

   edit_b = new QPushButton( this, "edit_b" );
   edit_b->setText( tr( "Open" ) );
   Layout2->addWidget( edit_b );

   new_b = new QPushButton( this, "new_b" );
   new_b->setText( tr( "New" ) );
   new_b->hide();
   Layout2->addWidget( new_b );

   delete_b = new QPushButton( this, "delete_b" );
   delete_b->setText( tr( "Delete" ) );
   Layout2->addWidget( delete_b );

   // signals and slots connections
   connect( up_b,     SIGNAL( clicked() ), this, SLOT( upClicked() ) );
   connect( down_b,   SIGNAL( clicked() ), this, SLOT( downClicked() ) );
   connect( edit_b,   SIGNAL( clicked() ), this, SLOT( editClicked() ) );
   connect( new_b,    SIGNAL( clicked() ), this, SLOT( newClicked() ) );
   connect( delete_b, SIGNAL( clicked() ), this, SLOT( deleteClicked() ) );
   connect( list_v,   SIGNAL( doubleClicked( QListViewItem *) ), 
         this, SLOT( editClicked(QListViewItem *) ) );

   m_last = NULL;

   loadLogList();

   show();
}

ErrList::~ErrList()
{
}

void ErrList::addColumn(const QString & title, int width = -1)
{
   list_v->addColumn( title, width );
}

void ErrList::addItem( QStringList & list )
{
   QListViewItem *item = new QListViewItem(list_v, m_last);

   int n = 0;
   for ( QStringList::Iterator it = list.begin(); it != list.end(); it++ )
   {
      item->setText(n, *it);
      n++;
   }

   m_last = item;
}

void ErrList::getItem(int n, QStringList & list )
{
   QListViewItem *item = list_v->firstChild();

   int cur = 0;
   while ( item && (cur < n) )
   {
      item = item->nextSibling();
      cur++;
   }

   list.clear();
   if ( item )
   {
      for ( int t = 0; t < list_v->columns(); t++ )
      {
         list << item->text(t);
      }
   }
}

int ErrList::items(void)
{
   int cur = 0;
   QListViewItem *item = list_v->firstChild();

   while ( item )
   {
      cur++;
      item = item->nextSibling();
   }

   return cur;
}

void ErrList::deleteClicked()
{
   QListViewItem *item = list_v->currentItem();
   if ( list_v->currentItem() )
   {
      if ( item )
      {
         QString file = tr(prefs.homeDir()) + tr("/.bronco/") + item->text(0);

         if ( unlink(file.latin1()) != 0 )
         {
            QMessageBox::warning(this, "Bronco", tr("Could not unlink ") + file + tr("\n") + strerror(errno));
         }
      }

      delete item;
   }
}

void ErrList::editClicked()
{
   editClicked( list_v->currentItem() );
}

void ErrList::editClicked( QListViewItem *item )
{
   if ( item )
   {
      QString file = tr(prefs.homeDir()) + tr("/.bronco/") + item->text(0);

      FILE *fp;

      if ( (fp = fopen(file.latin1(), "r")) != NULL )
      {
         char input[1024];

         ErrText *text = new ErrText(item->text(0));

         while ( fgets(input, sizeof(input), fp) )
         {
            int n = strlen(input) - 1;
            while ( n >= 0 && input[n] == '\n')
               input[n] = '\0';
            text->appendText(input);
         }
         fclose(fp);
      }
   }
}

void ErrList::newClicked()
{
}

void ErrList::downClicked()
{
   QListViewItem *item  = list_v->currentItem();
   QListViewItem *after = (item) ? item->nextSibling() : NULL;

   if ( item && after )
   {
      QListViewItem *create = new QListViewItem(list_v, after);
      list_v->setCurrentItem( create );
      list_v->ensureItemVisible( create );

      for( int t = 0; t < list_v->columns(); t++ )
      {
         create->setText(t, item->text(t) );
      }

      delete item;
   }
}

void ErrList::upClicked()
{
   QListViewItem *item  = list_v->currentItem();
   QListViewItem *after = (item) ? item->itemAbove() : NULL;

   if ( after )
      after = after->itemAbove();
   else
      after = NULL;

   if ( item )
   {
      QListViewItem *create = new QListViewItem(list_v, after);
      list_v->setCurrentItem( create );
      list_v->ensureItemVisible( create );

      for( int t = 0; t < list_v->columns(); t++ )
      {
         create->setText(t, item->text(t) );
      }

      delete item;
   }
}

void ErrList::clear(void)
{
   list_v->clear();
}

void ErrList::loadLogList(void)
{
   DIR *dp;
   struct dirent *d;

   QListViewItem *item = NULL;
   QString dirname = tr(prefs.homeDir()) + tr("/.bronco");
   QString file;

   if((dp = opendir(dirname)) != NULL){
      while((d = readdir(dp)) != NULL){
         if( strncmp("errorlog", d->d_name, 8) != 0) 
            continue;

         item = new QListViewItem(list_v, item, d->d_name);
      }

      closedir(dp);
   }
}
