/* execbr.cpp - Wrapper to fork, exec, pipe, ect..
 * 
 *  Bronco File Manager
 * 
 *  Copyright (c) 2000, 2001 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include "execbr.h"
#include "bronco.h"
#include "prefs.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

BroncoExec::BroncoExec()
{
   m_stdout     = NULL;
   m_stderr     = NULL;
   m_stdin      = NULL;

   m_pipeStdout = false;
   m_pipeStderr = false;
   m_pipeStdin  = false;

   m_wait       = false;

   m_childPid   = 0;
   m_parentPid  = 0;
}

BroncoExec::~BroncoExec()
{
   if ( m_stdout )
   {
      fclose(m_stdout);
      m_stdout = NULL;
   }
   if ( m_stderr )
   {
      fclose(m_stderr);
      m_stderr = NULL;
   }
   if ( m_stdin )
   {
      fclose(m_stdin);
      m_stdin = NULL;
   }
}

void BroncoExec::setWait(bool o)
{
   m_wait = o;
}

void BroncoExec::setPipeStdout(bool o)
{
   m_pipeStdout = o;
}

void BroncoExec::setPipeStderr(bool o)
{
   m_pipeStderr = o;
}

void BroncoExec::setPipeStdin(bool o)
{
   m_pipeStdin = o;
}

FILE * BroncoExec::getPipeStdout(void)
{
   return m_stdout;
}

FILE * BroncoExec::getPipeStderr(void)
{
   return m_stderr;
}

FILE * BroncoExec::getPipeStdin(void)
{
   return m_stdin;
}

int BroncoExec::exec(const char *command, const char * args[])
{
   char * exec_args[MAXARGS];
   char * program = NULL;
   int current_arg = 0;

   int e = 0;
   FILE *fp = NULL;

   int stdout_fd[2];
   int stderr_fd[2];
   int stdin_fd[2];

   if ( m_pipeStdout )
      pipe(stdout_fd);
   if ( m_pipeStderr )
      pipe(stderr_fd);
   if ( m_pipeStdin )
      pipe(stdin_fd);

   int exit_status = 0;

   int r = fork();

   switch ( r )
   {
      case 0:   /* Child */
         m_childPid  = getpid();
         m_parentPid = getppid();

         program = strdup(command);

         current_arg = 0;
         exec_args[ current_arg ] = strtok(program, " ");

         /* Spit our first argument up into its component parts.
          * This allows a lazy programmer (that's me!) to put all my
          * arguments in one string and have BroncoExec break them
          * up for me.
          */
         if ( exec_args[current_arg] )
         {
            do
            {
               current_arg++;
            } while( current_arg < (MAXARGS - 1) && NULL != (exec_args[current_arg] = strtok(NULL, " ")) );
         }

         if ( args )
         {
            for ( int n = 0; args[n] && current_arg < (MAXARGS - 1) ; args++ )
            {
               exec_args[current_arg] = strdup(args[n]);
               current_arg++;
            }
         }

         exec_args[current_arg] = NULL;

         /* Create requested pipes */
         if ( m_pipeStdout )
         {
            close( stdout_fd[0] );
            dup2( stdout_fd[1], fileno(stdout) );
         }

         if ( m_pipeStderr )
         {
            close( stderr_fd[0] );
            dup2( stderr_fd[1], fileno(stderr) );
         }

         if ( m_pipeStdin )
         {
            close( stdin_fd[1] );
            dup2( stdin_fd[0], fileno(stdin) );
         }

         /* cross our fingers and exec the program */

         execvp(program, exec_args);

         /* No need to free memory since we're exiting 
          * If I wanted to free memory I'd have to free program
          * and the exec_args that were strdup'd... which is not
          * all of them.  A marker to tell us where we started
          * strdup'ing would be sufficient
          */

         e = errno;

         fp = fopen(prefs.getFifoFileName(), "w");

         if ( fp )
         {
            fprintf(fp, "11 %d %s: %s\n", getpid(), program, strerror(errno));
         }
         else
         {
            fprintf(stderr, "Could not open %s: %s\n", prefs.getFifoFileName(), strerror(errno));
            fprintf(stderr, "%s: %s\n", program, strerror(errno));
         }
         exit(-1);
         return -1;
         break;

      default:  /* Parent */
         m_childPid  = r;
         m_parentPid = getpid();

         if ( m_pipeStdout )
         {
            close( stdout_fd[1] );
            m_stdout = fdopen(stdout_fd[0], "r");
         }

         if ( m_pipeStderr )
         {
            close( stderr_fd[1] );
            m_stderr = fdopen(stderr_fd[0], "r");
         }

         if ( m_pipeStdin )
         {
            close( stdin_fd[0] );
            m_stdin = fdopen(stdin_fd[1], "w");
         }

         if ( m_wait )
         {
            waitpid( m_childPid, &exit_status, 0);
         }
         else
         {
            return 0;
         }

         break;

      case -1:  /* Error */
         return -1;
         break;

   }
   return 0;
}

int BroncoExec::exec(const char *command)
{
   return exec(command, (const char **) NULL);
}

int BroncoExec::exec(const char *command, const char * file)
{
   const char *args[2];
   args[0] = file;
   args[1] = NULL;

   return exec(command, args);
}

int BroncoExec::wait(void)
{
   int exit_status = 0;

   waitpid ( m_childPid, &exit_status, 0);
   if ( WIFEXITED(exit_status) )
   {
      return( WEXITSTATUS(exit_status) );
   }
   else
   {
      return -1;
   }
}

int BroncoExec::killChild(void)
{
   kill(m_childPid, SIGTERM);
   return 0;
}

