/* filewin.h - Class to handle file operations
 * 
 *  Bronco File Manager
 * 
 *  Copyright (c) 2000, 2001 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include "filewin.h"
#include <qpushbutton.h>
#include <stdarg.h>

/* This class is a lot of "fun".  Originally the file operation window
 * was designed to have a progress bar and a status line.  Unfortunately,
 * Qt would not refresh the window while the operation was in progress.
 * There were two alternatives, do non-blocking writes, return to the
 * Qt event loop, and use callbacks to continue the file operations--or
 * fork and do the operations in a separate process.  I attemped the
 * first option for a while, and finally settled on the second.
 *
 * This class reflects that transition.  This class is still a progress
 * dialog.  It also is the forked process that handles copying, moving,
 * and deleting.  File operation progress is reported back to the main
 * program through a FIFO.  This is why Bronco has the FIFO, and it
 * was the beginning of the Helper API.
 *
 * Anyway... this class is a mess.
 */

/* Unique number for each file progress window */
int FileWin::s_nextFileWin = 1;
FileWinHash FileWin::s_winHash;

FileWin::FileWin(QWidget *parent, const char *name) : QWidget(parent, name, WDestructiveClose)
{
   setCaption("Bronco");
   QVBoxLayout *vbox = new QVBoxLayout(this, 5);
   curfile = new QLineEdit(this, "progress");
   pbar    = new QProgressBar(this, "progress");

   curfile->setReadOnly(TRUE);

   vbox->addWidget(curfile);
   vbox->addWidget(pbar);

   QSpacerItem * spacer       = new QSpacerItem( 20, 20, QSizePolicy::Minimum, QSizePolicy::Expanding );
   QHBoxLayout * buttonLayout = new QHBoxLayout( vbox );

   cancel_b = new QPushButton("Cancel", this, "cancel_b");

   buttonLayout->addItem(spacer);
   buttonLayout->addWidget(cancel_b);

   m_fifofp  = NULL;
   m_errorFp = NULL;
   overwrite = 1;

   m_winNumber = s_nextFileWin++;
   s_winHash[m_winNumber] = this;

   connect(cancel_b, SIGNAL(clicked()), this, SLOT(cancelPressed()));

   pbar->setProgress(0);
   percent = 0;

   this->resize(350, 125);
}

FileWin::~FileWin()
{
}

int FileWin::setDest(const char *d)
{
   snprintf(dest, sizeof(dest), "%s", d);

   int n = strlen(dest) - 1;
   while ( n >= 0 && dest[n] == '/' )
      dest[n--] = '\0';

   return(0);
}

/* This is where the copy starts, the fork happens in this function */
int FileWin::clipCopy(void)
{
   char *fname;
   char srcdir[1024];
   char destdir[1024];
   char filename[1024];
   int exists;
   int error;
   struct stat sbuf;

   setCaption("Bronco - Copy Status");

   if(prefs.confCopy()) {
      switch( QMessageBox::warning( this, "Copy confirmation",
               QString("Do you wish to copy selected \n") +
               QString("files to the current directory?\n"),
               QMessageBox::Yes    | QMessageBox::Default, QMessageBox::No     | QMessageBox::Escape )) {
         case QMessageBox::Yes:
            break;
         case QMessageBox::No:
            return(0);
            break;
      }
   }

//   curfile->setText("computing size...");

   bdone  = 0;
   btotal = 0;
   exists = 0;
   error  = 0;
   overwrite = 1;
   if(clipDir.first(filename) )
   {
      do
      {
         fname = rindex(filename, '/');

         if ( fname )
            fname++;
         else
            fname = filename;

         snprintf(srcdir,  sizeof(srcdir),  "%s/%s", clip_dir, filename);
         snprintf(destdir, sizeof(destdir), "%s/%s", dest,     fname);

         if(strcmp(srcdir, destdir) == 0)
         {
            QMessageBox::warning( this, "Files are the same", "Source and destination are the same.", QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
            return(0);
         }

         if(strncmp(srcdir, destdir, strlen(srcdir)) == 0)
         {
            QMessageBox::warning( this, "Copying into Subdirectory", "You cannot copy a directory into\na subdirectory of itself.", QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
            return(0);
         }

         if(lstat(destdir, &sbuf) >= 0)
            exists++;

         btotal += compute_blocks(srcdir);
      } while( clipDir.next(filename) );
   }

   if(exists)
   {
      if(prefs.confOverWrite()) {
         switch( QMessageBox::warning( this, "Over-write confirmation",
                  "Some files selected to copy exist\n"
                  "at the target location.  Do you\n"
                  "wish to over-write?", QMessageBox::Yes    | QMessageBox::Default, QMessageBox::No, QMessageBox::Cancel | QMessageBox::Escape )) {
            case QMessageBox::Yes:
               break;
            case QMessageBox::Cancel:
               return(0);
               break;
            case QMessageBox::No:
               overwrite = 0;
               break;
         }
      }
   }

   this->show();
   m_pid = fork();

   switch ( m_pid )
   {
      default:
         return 1;
         break;
      case 0:
         break;
      case -1:
         QMessageBox::warning( this, "fork()", strerror(error), QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
         return 0;
         break;
   }

   signal(SIGINT, FileWin::sigCancel);

   m_fifofp = fopen(prefs.getFifoFileName(), "w");

   if ( m_fifofp == NULL)
   {
      errorlog("fifo: %s\n", strerror(errno));
      exit(0);
   }

   if(clipDir.first(filename) )
   {
      do
      {
         fname = rindex(filename, '/');

         if ( fname )
            fname++;
         else
            fname = filename;

         snprintf(srcdir,  sizeof(srcdir),  "%s/%s", clip_dir, filename);
         snprintf(destdir, sizeof(destdir), "%s/%s", dest,     fname);

         fprintf(m_fifofp, "82 %d %s\n", m_winNumber, srcdir);
         fflush(m_fifofp);

         if(copy(srcdir, destdir) == 0)
            error = errno;
      } while( clipDir.next(filename) );
   }
   
   if(error)
   {
//      QMessageBox::warning( this, "Copy failed", strerror(error), QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
      operationDone();
   }
   else
   {
      operationDone();
   }
   return 0;
}

/* This is where the move starts, the fork happens in this function */
int FileWin::clipMove(void)
{
   char srcdir[1024];
   char destdir[1024];
   char filename[1024];
   int exists;
   int error;
   struct stat sbuf;

   setCaption("Bronco - Move Status");

   if(prefs.confMove()) {
      switch( QMessageBox::warning( this, "Move confirmation",
               QString("Do you wish to move selected \n") +
               QString("files to the current directory?\n"),
               QMessageBox::Yes    | QMessageBox::Default, QMessageBox::No     | QMessageBox::Escape )) {
         case QMessageBox::Yes:
            break;
         case QMessageBox::No:
            return(0);
            break;
      }
   }
//   curfile->setText("computing size...");

   bdone  = 0;
   btotal = 0;
   exists = 0;
   error  = 0;
   overwrite = 1;
   if(clipDir.first(filename) )
   {
      do
      {
         snprintf(srcdir,  sizeof(srcdir),  "%s/%s", clip_dir, filename);
         snprintf(destdir, sizeof(destdir), "%s/%s", dest,     filename);

         if(strcmp(srcdir, destdir) == 0)
         {
            QMessageBox::warning( this, "Files are the same!", "Source and destination are the same.", QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
            return(0);
         }

         if(strncmp(srcdir, destdir, strlen(srcdir)) == 0)
         {
            QMessageBox::warning( this, "Moving into Subdirectory", "You cannot move a directory into\na subdirectory of itself.", QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
            return(0);
         }

         if(lstat(destdir, &sbuf) >= 0)
            exists++;

         btotal += compute_blocks(srcdir);
      } while( clipDir.next(filename) );
   }

   if(exists)
   {
      if(prefs.confOverWrite()) {
         switch( QMessageBox::warning( this, "Over-write confirmation",
                  "Some files selected to move exist\n"
                  "at the target location.  Do you\n"
                  "wish to continue?", QMessageBox::Yes    | QMessageBox::Default, QMessageBox::No     | QMessageBox::Escape )) {
            case QMessageBox::Yes:
               break;
            case QMessageBox::No:
               return(0);
               break;
         }
      }
   }

   this->show();
   m_pid = fork();

   switch ( m_pid )
   {
      default:
         return 1;
         break;
      case 0:
         break;
      case -1:
         QMessageBox::warning( this, "fork()", strerror(error), QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
         return 0;
         break;
   }

   m_fifofp = fopen(prefs.getFifoFileName(), "w");

   if ( m_fifofp == NULL)
   {
      errorlog("fifo: %s: %s\n", strerror(errno));
      exit(0);
   }

   if(clipDir.first(filename) )
   {
      do
      {
         snprintf(srcdir,  sizeof(srcdir),  "%s/%s", clip_dir, filename);
         snprintf(destdir, sizeof(destdir), "%s/%s", dest,     filename);
//         curfile->setText(srcdir);
         fprintf(m_fifofp, "82 %d %s\n", m_winNumber, srcdir);
         fflush(m_fifofp);

         if(move(srcdir, destdir) == 0)
            error = errno;
      } while( clipDir.next(filename) );
   }

   if(error)
   {
//      QMessageBox::warning( this, "Copy failed", strerror(error), QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
      operationDone();
   }
   else
   {
      operationDone();
   }
   return 0;
}

/* This is where the delete starts, the fork happens in this function */
int FileWin::clipDelete(void)
{
   char srcdir[1024];
   char filename[1024];
   int error = 0;

   setCaption("Bronco - Delete Status");

//   curfile->setText("computing size...");

   if(prefs.confDelete()) {
      switch( QMessageBox::warning( this, "Delete confirmation",
               "Are you sure you wish to delete\n"
               "the selected files?\n", QMessageBox::Yes | QMessageBox::Default, QMessageBox::No  | QMessageBox::Escape )) {
         case QMessageBox::Yes:
            break;
         case QMessageBox::No:
            return(0);
            break;
      }
   }

   this->show();
   m_pid = fork();

   switch ( m_pid )
   {
      default:
         return 1;
         break;
      case 0:
         break;
      case -1:
         QMessageBox::warning( this, "fork()", strerror(error), QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
         return 0;
         break;
   }

   m_fifofp = fopen(prefs.getFifoFileName(), "w");

   if ( m_fifofp == NULL)
   {
      errorlog("fifo: %s\n", strerror(errno));
      exit(0);
   }

   bdone  = 0;
   btotal = 0;
   error  = 0;
   if(clipDir.first(filename) )
   {
      do
      {
         snprintf(srcdir, sizeof(srcdir), "%s/%s", clip_dir, filename);
         btotal += compute_blocks(srcdir);
      } while( clipDir.next(filename) );
   }

   if(clipDir.first(filename) )
   {
      do
      {
         snprintf(srcdir, sizeof(srcdir), "%s/%s", clip_dir, filename);
//         curfile->setText(srcdir);
         fprintf(m_fifofp, "82 %d %s\n", m_winNumber, srcdir);
         fflush(m_fifofp);

         if(del(srcdir) == 0)
            error = errno;
      } while( clipDir.next(filename) );
   }

   if(error)
   {
//      QMessageBox::warning( this, "Delete failed", strerror(error), QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
      operationDone();
   }
   else
   {
      operationDone();
   }
   return 0;
}

/* t is the number of new blocks that have been finished.
 * We add that to blocks complete, and report a percentage back
 * to the main process via the FIFO */
void FileWin::blocks_done(long int t)
{
   bdone += t;
   if ( btotal )
   {
      long new_percent = (bdone * 100 / btotal);
      if ( new_percent > percent)
      {
         percent = new_percent;
         fprintf(m_fifofp, "80 %d %ld\n", m_winNumber, new_percent );
      }
   }
   fflush( m_fifofp );
   return;
}

/* Get total number of blocks for this operation */
long int FileWin::compute_blocks(const char *f1)
{
   struct stat sbuf;
   struct dirent *dbuf;
   DIR *dp;
   char filename[1024];
   long int total = 0;

   if( lstat(f1, &sbuf) == 0 )
   {
      if( S_ISDIR(sbuf.st_mode) )
      {
         if( ( dp = opendir(f1) ) != NULL)
         {
            while( (dbuf = readdir(dp)) != NULL )
            {
               if(strcmp(".", dbuf->d_name) != 0 && strcmp("..", dbuf->d_name) != 0)
               {
                  snprintf(filename, sizeof(filename), "%s/%s", f1, dbuf->d_name);
                  total += compute_blocks(filename);
               }
            }
            closedir(dp);

            total += sbuf.st_blocks;
            return(total);
         }
         else
         {
            return(0);
         }
      }
      else
      {
         return(sbuf.st_blocks);
      }
      return(1);
   }
   else
   {
      return(0);
   }
}

/* Deletes recursively.  The delete begins in clipDelete */
int FileWin::del(const char *f1)
{
   struct stat sbuf;
   struct dirent *dbuf;
   DIR *dp;
   char filename[1024];

   if( lstat(f1, &sbuf) == 0 )
   {
      if( S_ISDIR(sbuf.st_mode) )
      {

         /* Delete directory */
         
         if( ( dp = opendir(f1) ) != NULL)
         {
            while( (dbuf = readdir(dp)) != NULL )
            {
               if(strcmp(".", dbuf->d_name) != 0 && strcmp("..", dbuf->d_name) != 0)
               {
                  snprintf(filename, sizeof(filename), "%s/%s", f1, dbuf->d_name);
                  del(filename);
               }
            }
            closedir(dp);

            if( rmdir(f1) == 0 )
            {
               blocks_done(sbuf.st_blocks);
               return(1);
            }
            else
            {
               errorlog("rmdir: %s: %s\n", f1, strerror(errno));
               return(0);
            }
         }
         else
         {
            errorlog("opendir: %s: %s\n", f1, strerror(errno));
            return(0);
         }
      }
      else
      {

         /* Delete file */
         
         if( unlink(f1) == 0 )
         {
            blocks_done(sbuf.st_blocks);
            return(1);
         }
         else
         {
            errorlog("delete: %s: %s\n", f1, strerror(errno));
            return(0);
         }
      }
      return(1);
   }
   else
   {
      errorlog("stat: %s: %s\n", f1, strerror(errno));
      return(0);
   }
}

/* Copies a file.  The copy begins in clipCopy.
 * clipCopy calls copy, which calls this function to copy files */
int FileWin::copy_file(const char *f1, const char *f2, mode_t st_mode)
{
   FILE *fp1, *fp2;
   char buf[4096];
   struct stat sbuf;
   int n;

   /* If file exists, and no over-write, return success */
   if((overwrite == 0) && (lstat(f2, &sbuf) >= 0))
      return(1);  /* return success, do not over-write */

   if( (fp1 = fopen(f1, "r") ) != NULL )
   {
      if( (fp2 = fopen(f2, "w") ) != NULL )
      {
         while( (n = fread(buf, sizeof(char), sizeof(buf), fp1)) > 0)
         {
            if( fwrite( buf, sizeof(char), n, fp2 ) == 0 )
            {
               errorlog("write: %s: %s\n", f2, strerror(errno));
               fclose(fp1);
               fclose(fp2);
               return(0);
            }
            blocks_done( (n / 512) + (n % 512 ? 1 : 0) );
         }

         if(n < 0)
         {
            errorlog("read: %s: %s\n", f1, strerror(errno));
            return(0);
         }

         fclose(fp2);
         fclose(fp1);

         chmod(f2, (st_mode & (S_IRWXU | S_IRWXG | S_IRWXO)) | S_IWUSR );
         return(1);
      }
      else
      {
         errorlog("open: %s: %s\n", f2, strerror(errno));
         fclose(fp1);
         return(0);
      }
   }
   else
   {
      errorlog("open: %s: %s\n", f1, strerror(errno));
      return(0);
   }
}

/* Copies files and directories recursively.  The copy begins in 
 * clipCopy. clipCopy calls this function, which copies directories
 * and calls copy_file to copy files */
int FileWin::copy(const char *f1, const char *f2)
{
   struct stat sbuf;
   struct dirent *dbuf;
   DIR *dp;
   char src[1024];
   char dest[1024];

   if( lstat(f1, &sbuf) == 0 )
   {
      if( S_ISDIR(sbuf.st_mode) )
      {

         /* Copy directory */
         
         if( ( dp = opendir(f1) ) != NULL)
         {
            if(mkdir(f2, (sbuf.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO)) | S_IWUSR) == 0 || errno == EEXIST)
            {
               while( (dbuf = readdir(dp)) != NULL )
               {
                  if(strcmp(".", dbuf->d_name) != 0 && strcmp("..", dbuf->d_name) != 0)
                  {
                     snprintf(src,  sizeof(src),  "%s/%s", f1, dbuf->d_name);
                     snprintf(dest, sizeof(dest), "%s/%s", f2, dbuf->d_name);

                     if(copy(src, dest) == 0)
                     {
                        /* this only stops the copy if 
                         * "Continue on error" is disabled.
                         * Otherwise, copy (copy_file) returns TRUE */
                        closedir(dp);
                        return(0);
                     }
                  }
               }
               closedir(dp);
               blocks_done(sbuf.st_blocks);
               chmod(f2, (sbuf.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO)) | S_IWUSR);
               return(1);
            }
            else
            {
               errorlog("mkdir: %s: %s\n", f2, strerror(errno));
               closedir(dp);
               return(0);
            }
         }
         else
         {
            errorlog("open: %s: %s\n", f1, strerror(errno));
            return(0);
         }
      }
      else if( S_ISFIFO(sbuf.st_mode) )
      {
         if( mkfifo(f2, sbuf.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO) ) )
         {
            blocks_done(sbuf.st_blocks);
            return(1);
         }
         else
         {
            errorlog("mkfifo: %s: %s\n", f2, strerror(errno));
            return(0);
         }
      }
      else if( S_ISLNK(sbuf.st_mode) )
      {
         int n;
         if((n = readlink(f1, src, sizeof(src))) >= 0)
         {
            src[n] = '\0';
            if(symlink(src, f2) == 0)
            {
               blocks_done(sbuf.st_blocks);
               return(1);
            }
            else
            {
               errorlog("symlink: %s: %s\n", f2, strerror(errno));
               return(0);
            }
         }
         else
         {
            errorlog("readlink: %s: %s\n", f2, strerror(errno));
            return(0);
         }
      }
      else if ( S_ISREG(sbuf.st_mode) )
      {

         /* Copy file */
         
         if( copy_file(f1, f2, sbuf.st_mode) )
         {
//            blocks_done(sbuf.st_blocks);
            return(1);
         }
         else
         {
            return(0);
         }
      }
      else
      {
         /* Ignore special files */
         errorlog("ignoring special file: %s\n", f2);
         return(0);
      }
      
      return(1);
   }
   else
   {
      errorlog("stat: %s: %s\n", f1, strerror(errno));
      return(0);
   }
}

/* Moves files recursively, the move begins in clipMove */
int FileWin::move(const char *f1, const char *f2)
{
   if(rename(f1, f2) == 0 )
   {
      return(1);
   }
   else if(errno == EXDEV)
   {
      if(copy(f1, f2))
      {
         if( del(f1) )
         {
            return(1);
         }
         else
         {
            int error = errno;
            del(f2);
            errorlog("move failed\ncopy move failed: %s\n", strerror(error));
            return(0);
         }
      }
      else
      {
         return(0);
      }
   }
   else
   {
      if(errno == ENOTEMPTY)
      {
         if(copy(f1, f2) == 1)
         {
            del(f1);
            return(1);
         }
      }

      errorlog("directory not empty\ncopy move failed: %s\n", strerror(errno));
      return (0);
   }
}

void FileWin::sigYes(int s)
{
   m_answer = 2;
}

void FileWin::sigNo(int s)
{
   m_answer = 1;
}

/* User clicked 'cancel' in the progress dialog.
 * We got caught SIGINT, time to die */
void FileWin::sigCancel(int s)
{
   exit(0);
}

/* Ask main process to close the progress dialog and refresh
 * the destination directory */
void FileWin::operationDone(void)
{
   fprintf(m_fifofp, "81 %d\n", m_winNumber);
   fprintf(m_fifofp, "51 %d %s\n", getpid(), dest);
   exit(0);
}

/* get the window pointer for the window number */
FileWin * FileWin::getWinNumber(int w)
{
   return s_winHash[w];
}

void FileWin::setProgress(int p)
{
   pbar->setProgress(p);
}

void FileWin::setCurrentFile(const char *s)
{
   curfile->setText(s);
}

/* User clicked 'cancel'.  Send SIGINT to child */
void FileWin::cancelPressed(void)
{
   s_winHash.erase(m_winNumber);
   if ( m_pid > 0 )
   {
      kill(m_pid, SIGINT);
   }
   hide();

   QTimer::singleShot(1000, this, SLOT(goAway()));
}

/* close() is protected.  How lame is that? */
void FileWin::goAway(void)
{
   close();
}

/* printf style error loging.  If the log isn't open for writing,
 * then we open it.  The user can see this log after we are done
 */
void FileWin::errorlog(const char *fmt, ...)
{
   va_list ap;
   va_start(ap, fmt);

   if ( m_errorFp == NULL)
   {
      QString file;
      file.sprintf("%s/.bronco/errorlog%d", prefs.homeDir(), getpid());
      m_errorFp = fopen(file.latin1(), "w");
   }

   if ( m_errorFp )
   {
      vfprintf(m_errorFp, fmt, ap);
   }
}

