/* findwin.cpp - Window for searching
 * 
 *  Bronco File Manager
 * 
 *  Copyright (c) 2000, 2001 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include <sys/wait.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <qpushbutton.h>
#include "bronco.h"
#include "find.h"
#include "findwin.h"
#include "prefs.h"
#include "dirlist.h"

FILE * startFind ( const char *dir, const char *glob, FindWin *win)
{
   int fd[2];
   int pid;

   pipe( fd );

   pid = fork();

   if ( pid == 0 )
   {
      close( fd[0] );

      FILE * fp = fdopen(fd[1], "w");

      findFile(dir, glob, NULL, fp);

      fclose(fp);
      exit(0);
   }
   else
   {
      close( fd[1] );

      FILE * fp = fdopen(fd[0], "r");

      win->setFindPid( pid );

      return fp;

   }

   return NULL;
}

FindWin::FindWin(QWidget *parent, const char *name) : QWidget(parent, name, WDestructiveClose){
   int x, y, p;
   QAccel *accel;

   setMinimumSize(550, 350);
   setCaption("Find");

   list = new DirList(this, "");
   list->shortMenu();
   list->setSorting(-1, FALSE);

   /* Initilization */

   /* labels */
   what_l    = new QLabel("Search For",         this, "");
   where_l   = new QLabel("Starting In",        this, "");

   /* values */
   what_v  = new QLineEdit(this, "");
   where_v = new QLineEdit(this, "");

   case_v    = new QCheckBox("&Case Sensitive", this, "");
   recurse_v = new QCheckBox("&Recursive", this, "");
   regex_v   = new QCheckBox("Regular E&xpression", this, "");

   p = prefs.getFindPrefs();
   case_v->setChecked(    (p & FIND_CASE) ? TRUE : FALSE);
   regex_v->setChecked(   (p & FIND_REG)  ? TRUE : FALSE);
   recurse_v->setChecked( (p & FIND_REC)  ? TRUE : FALSE);
   
   search_b  = new QPushButton("Find", this, "");
   search_b->setDefault(true);

   GrepWinLayout = new QVBoxLayout( this ); 
   GrepWinLayout->setSpacing( 6 );
   GrepWinLayout->setMargin( 11 );

   Layout2 = new QGridLayout; 
   Layout2->setSpacing( 6 );
   Layout2->setMargin( 0 );

   Layout2->addWidget( what_l, 0, 0 );

   Layout2->addWidget( recurse_v, 0, 2 );

   Layout2->addWidget( case_v, 0, 3 );

   Layout1 = new QHBoxLayout; 
   Layout1->setSpacing( 6 );
   Layout1->setMargin( 0 );

   Layout1->addWidget( search_b );

   cancel_b = new QPushButton( this, "cancel_b" );
   cancel_b->setEnabled(false);
   cancel_b->setText( tr( "Cancel" ) );
   Layout1->addWidget( cancel_b );

   Layout2->addLayout( Layout1, 1, 3 );

   Layout2->addWidget( where_v, 1, 1 );

   Layout2->addWidget( where_l, 1, 0 );

   Layout2->addWidget( what_v, 0, 1 );

   Layout2->addWidget( regex_v, 1, 2 );
   GrepWinLayout->addLayout( Layout2 );

   GrepWinLayout->addWidget( list );

   read_t = new QTimer(this, "");

   connect(read_t,   SIGNAL(timeout()),       this,    SLOT(readTimeout()));
   connect(search_b, SIGNAL(clicked()),       this,    SLOT(doSearch()));
   connect(cancel_b, SIGNAL(clicked()),       this,    SLOT(cancelClicked()));
   connect(what_v,   SIGNAL(returnPressed()), this,    SLOT(doSearch()));
   connect(where_v,  SIGNAL(returnPressed()), this,    SLOT(doSearch()));
   connect(list,     SIGNAL(directoryChanged(const QString &)), 
         this, SLOT(updateSearchIn(const QString &)));

   what_v->setFocus();

   accel = new QAccel(this);
   if(parent == NULL) accel->connectItem(accel->insertItem(Key_W+CTRL),       this, SLOT(close()));

   if(parent == NULL){
      prefs.getFindSize(&x, &y);
      this->resize(x, y);
   }

   find_fp = NULL;
   find_pid = 0;

   show();
}

FindWin::~FindWin(){
}

void FindWin::resizeEvent(QResizeEvent *e){
   list->resize(this->width() - 10, this->height() - 60);

   if(this->parent() == NULL)
   {
      prefs.setFindSize(this->width(), this->height());
      prefs.save();
   }
}

void FindWin::setDir(const char *d){
   where_v->setText(d);
   return;
}

void FindWin::doSearch(void){
   char buffer[4096];
   int s;

   char * glob = strdup(what_v->text().latin1());
   char * dir = strdup(where_v->text().latin1());

   snprintf(buffer, sizeof(buffer), "find %s %s", glob, dir);

   s  = 0;
   s |= regex_v->isChecked()   ? FIND_REG : 0;
   s |= recurse_v->isChecked() ? FIND_REC : 0;
   s |= case_v->isChecked()    ? FIND_CASE : 0;

   list->clearList();
   
   find_fp = startFind(dir, glob, this);

   if ( find_fp ) 
   {
      cancel_b->setEnabled(true);
      search_b->setEnabled(false);
      list->loadDir(buffer);
      readTimeout();
   }

   free( glob );
   free( dir );

   return;
}

void FindWin::hide(void){
   int s;

   s  = 0;
   s |= regex_v->isChecked()   ? FIND_REG  : 0;
   s |= recurse_v->isChecked() ? FIND_REC  : 0;
   s |= case_v->isChecked()    ? FIND_CASE : 0;

   prefs.setFindPrefs(s);
   prefs.save();

   QWidget::hide();
}

void FindWin::updateSearchIn(const QString &dir)
{
   if ( dir[0] == QChar('/') )
   {
      where_v->setText(dir);
   }
}

void FindWin::readTimeout()
{
   char input[4096];
   fd_set read_fd;
   timeval tv;

   if ( find_fp )
   {
      bool done = false;
      
      while ( ! done )
      {
         FD_ZERO(&read_fd);

         FD_SET(fileno(find_fp), &read_fd);

         tv.tv_sec  = 0;
         tv.tv_usec = 0;

         select(fileno(find_fp)+1, &read_fd, NULL, NULL, &tv);

         if ( FD_ISSET(fileno(find_fp), &read_fd) )
         {
            if ( fgets(input, sizeof(input), find_fp) )
            {
               input[ strlen(input) - 1] = '\0';
               list->addFile( input );
            }
            else
            {
               list->doListUpdate();
               done = true;
               fclose( find_fp );
               find_fp  = NULL;
               find_pid = 0;

               cancel_b->setEnabled(false);
               search_b->setEnabled(true);
            }

         }
         else
         {
            list->doListUpdate();
            done = true;
            read_t->start(500, true);
         }
      }
   }
}

void FindWin::cancelClicked(void)
{
   read_t->stop();
   if ( find_fp)
   {
      kill(find_pid, SIGTERM);
      find_pid = 0;
      fclose(find_fp);
      find_fp = NULL;
   }

   cancel_b->setEnabled( false );
   search_b->setEnabled( true );
}

void FindWin::setFindPid(int pid)
{
   find_pid = pid;
}
