/* grepwin.cpp - Window for grepping
 * 
 *  Bronco File Manager
 * 
 *  Copyright (c) 2000, 2001 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include <unistd.h>
#include <ctype.h>
#include <dirent.h>
#include <qregexp.h>
#include <sys/types.h>
#include <sys/time.h>
#include "bronco.h"
#include "execbr.h"
#include "find.h"
#include "grepwin.h"
#include "openwith.h"
#include "prefs.h"

GrepWin::GrepWin(QWidget *parent, const char *name) : QWidget(parent, name, WDestructiveClose){
   int x, y, p;
   QAccel *accel;

   setMinimumSize(600, 350);
   setCaption("Grep");

   list = new QListView(this, "");
   list->setAllColumnsShowFocus(TRUE);

   list->addColumn("file", -1);
   list->addColumn("line", -1);
   list->addColumn("text", -1);

   /* Initilization */

   /* labels */
   what_l    = new QLabel("Grep For",         this, "");
   where_l   = new QLabel("Starting In",        this, "");

   what_v  = new QLineEdit(this, "");
   where_v = new QLineEdit(this, "");

   case_v     = new QCheckBox("&Case Sensitive", this, "");
   recurse_v  = new QCheckBox("Recursive", this, "");
   selected_v = new QCheckBox("Only Selected Files", this, "");

   p = prefs.getGrepPrefs();
   case_v->setChecked(    (p & FIND_CASE) ? TRUE : FALSE);
   recurse_v->setChecked( (p & FIND_REC)  ? TRUE : FALSE);
   
   search_b  = new QPushButton("Grep", this, "");
   search_b->setDefault(true);

   GrepWinLayout = new QVBoxLayout( this ); 
   GrepWinLayout->setSpacing( 6 );
   GrepWinLayout->setMargin( 11 );

   Layout2 = new QGridLayout; 
   Layout2->setSpacing( 6 );
   Layout2->setMargin( 0 );

   Layout2->addWidget( what_l, 0, 0 );

   Layout2->addWidget( recurse_v, 0, 2 );

   Layout2->addWidget( case_v, 0, 3 );

   Layout1 = new QHBoxLayout; 
   Layout1->setSpacing( 6 );
   Layout1->setMargin( 0 );

   Layout1->addWidget( search_b );

   cancel_b = new QPushButton( this, "cancel_b" );
   cancel_b->setText( tr( "Cancel" ) );
   cancel_b->setEnabled( false );
   Layout1->addWidget( cancel_b );

   Layout2->addLayout( Layout1, 1, 3 );

   Layout2->addWidget( where_v, 1, 1 );

   Layout2->addWidget( where_l, 1, 0 );

   Layout2->addWidget( what_v, 0, 1 );

   Layout2->addWidget( selected_v, 1, 2 );
   GrepWinLayout->addLayout( Layout2 );

   GrepWinLayout->addWidget( list );

   read_t = new QTimer(this, "");

   connect(read_t,     SIGNAL(timeout()),       this, SLOT(readTimeout()));
   connect(search_b,   SIGNAL(clicked()),       this, SLOT(doSearch()));
   connect(cancel_b,   SIGNAL(clicked()),       this, SLOT(cancelClicked()));
   connect(what_v,     SIGNAL(returnPressed()), this, SLOT(doSearch()));
   connect(where_v,    SIGNAL(returnPressed()), this, SLOT(doSearch()));
   connect(selected_v, SIGNAL(toggled(bool)),   this, SLOT(toggleWhere(bool)));
   connect(list,       SIGNAL(returnPressed(QListViewItem *)), 
         this, SLOT(openItem(QListViewItem *)));
   connect(list,       SIGNAL(doubleClicked(QListViewItem *)), 
         this, SLOT(openItem(QListViewItem *)));

   what_v->setFocus();

   accel = new QAccel(this);
   if(parent == NULL) accel->connectItem(accel->insertItem(Key_W+CTRL), 
         this, SLOT(close()));

   if(parent == NULL){
      prefs.getFindSize(&x, &y);
      this->resize(x, y);
   }

   m_lastItem = NULL;
   m_prog     = NULL;

//   show();
}

GrepWin::~GrepWin(){
}

void GrepWin::resizeEvent(QResizeEvent *e){
   list->resize(this->width() - 10, this->height() - 60);

   if(this->parent() == NULL)
   {
      prefs.setFindSize(this->width(), this->height());
      prefs.save();
   }
}

void GrepWin::setDir(const char *d){
   m_startDir = d;
   where_v->setText(d);
   return;
}

void GrepWin::addFile(const QString &s )
{
   m_selectedFiles << s;
}

void GrepWin::doSearch(void){
   const char *args[MAXARGS];
   int offset = 0;

   list->clear();

   if ( what_v->text().isEmpty() )
   {
      QMessageBox::warning(this, "Bronco", "Refusing to grep on an empty string");
      return;
   }


   if ( recurse_v->isChecked() )
      args[ offset++ ] = "-r";
   else
      args[ offset++ ] = "--directories=skip";

   if ( ! case_v->isChecked() )
      args[ offset++ ] = "-i";

   args[ offset++ ] = "-s";
   args[ offset++ ] = "-E";
   args[ offset++ ] = "-n";
   args[ offset++ ] = what_v->text().latin1();

   QStringList searchList;
   if ( selected_v->isChecked() )
   {
      m_lastSearchDir = m_startDir;
      chdir( m_startDir.latin1() );
      searchList = m_selectedFiles;
   }
   else
   {
      m_lastSearchDir = where_v->text();
      chdir( (where_v->text()).latin1() );

      DIR *dp;
      struct dirent *d;

      if ( (dp = opendir(where_v->text().latin1())) != NULL )
      {
         while( (d = readdir(dp)) != NULL )
         {
            if( (!strcmp(".", d->d_name) || !strcmp("..", d->d_name)) ) 
               continue;

            searchList << d->d_name;
         }
         closedir(dp);
      }
      else
      {
         QMessageBox::warning(this, "Bronco", "Could not read directory\n" + where_v->text());
         return;
      }
   }

   int fileCount = 0;
   for ( unsigned n = 0; n < searchList.count() && (offset < MAXARGS - 1); n++ )
   {
      args[ offset++ ] = searchList[n].latin1();
      fileCount++;
   }
   args[ offset ] = NULL;

   if ( fileCount <= 0 )
   {
      QMessageBox::warning(this, "Bronco", "No files to grep");
      return;
   }

   if ( m_prog )
   {
//      delete m_prog;
   }

   m_prog = new BroncoExec();
   m_prog->setPipeStdout( true );
   m_prog->exec("grep", args);

   grep_fp = m_prog->getPipeStdout();

   if ( grep_fp )
   {
      cancel_b->setEnabled( true );
      search_b->setEnabled( false );
      m_lastItem = NULL;

      readTimeout();
   }
   else
   {
      QMessageBox::warning( NULL, "Bronco", "Could not open pipe to grep", QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
   }

   return;
}

void GrepWin::show(void)
{
   cancel_b->setEnabled( false );
   search_b->setEnabled( true );

   if ( m_selectedFiles.count() >= 2 )
   {
      selected_v->setChecked(true);
   }
   else if ( m_selectedFiles.count() == 0 )
   {
      selected_v->setChecked(false);
      selected_v->setEnabled(false);
   }
   else
   {
      selected_v->setChecked(false);
   }
   QWidget::show();
}

void GrepWin::hide(void){
   int s;

   s  = 0;
//   s |= selected_v->isChecked()   ? FIND_SEL : 0;
   s |= recurse_v->isChecked() ? FIND_REC : 0;
   s |= case_v->isChecked()    ? FIND_CASE : 0;

   prefs.setGrepPrefs(s);
   prefs.save();

   if ( m_prog)
   {
      read_t->stop();
      m_prog->killChild();
   }
   QWidget::hide();
}

void GrepWin::updateSearchIn(const QString &dir)
{
   if ( dir[0] == QChar('/') )
   {
      where_v->setText(dir);
   }
}

void GrepWin::openItem( QListViewItem *item )
{
   const char *args[MAXARGS];
   BroncoExec prog;

   QString     command;
   QString     remainder;
   QString     current;
   QStringList arglist;

   chdir(m_lastSearchDir.latin1());

   command = prefs.getGrepEdit();
   command = command.simplifyWhiteSpace();

   if ( command.isEmpty() )
   {
      command = prefs.getCommand( item->text(0).latin1() );
      command = command.simplifyWhiteSpace();

      if ( command.isEmpty() )
      {
         OpenWith p(this, "");
         if ( p.exec() )
         {
            command = p.getInput();
            command = command.simplifyWhiteSpace();

            if ( command.isEmpty() )
            {
               QMessageBox::warning( NULL, "Bronco", "Open cancelled", QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape, 0);
               return;
            }
         }
         else
         {
            return;
         }
      }
   }

   remainder = command.right( command.length() - command.find( QRegExp("\\s") ) );
   command.truncate( command.find( QRegExp("\\s")) );

   remainder = remainder.stripWhiteSpace();

   while ( ! remainder.isEmpty() )
   {
      current = remainder;
      if ( remainder.find( QRegExp("\\s")) >= 0)
      {
         remainder = remainder.right( remainder.length() - remainder.find( QRegExp("\\s") ) );
         current.truncate( current.find( QRegExp("\\s")) );
      }
      else
      {
         remainder = "";
      }

      if ( current.contains("$L") )
      {
         current.replace( current.find("$L"), 2, item->text(0) );
         current.prepend( m_lastSearchDir + QString("/") );
      }

      if ( current.contains("$l") )
      {
         current.replace( current.find("$l"), 2, item->text(0).latin1() );
      }

      if ( current.contains("$F") )
      {
         current.replace( current.find("$F"), 2, item->text(0) );
         current.prepend( m_lastSearchDir + QString("/") );
      }

      if ( current.contains("$f") )
      {
         current.replace( current.find("$f"), 2, item->text(0).latin1() );
      }

      if ( current.contains("$n") )
      {
         current.replace( current.find("$n"), 2, item->text(1).latin1() );
      }

      if ( current.contains("$d") )
      {
         current.replace( current.find("$d"), 2, m_lastSearchDir );
      }

      arglist << current;

      remainder = remainder.stripWhiteSpace();
   }

   int n = 0;
   for ( QStringList::Iterator it = arglist.begin(); it != arglist.end(); it++ )
   {
      args[n] = (*it).latin1();
      n++;

      if ( n == MAXARGS - 1 )
         break;
   }
   args[n] = NULL;

   prog.exec(command.latin1(), args);
}

void GrepWin::toggleWhere(bool onlySelected )
{
   if ( m_selectedFiles.count() == 0 && onlySelected == true)
   {
      selected_v->setChecked(false);
   }
   else
   {
      if ( onlySelected )
      {
         where_v->setEnabled(false);
      }
      else
      {
         where_v->setEnabled(true);
      }
   }
}

void GrepWin::readTimeout()
{
   char input[4096];
   fd_set read_fd;
   timeval tv;

   if ( grep_fp )
   {
      bool done = false;
      
      while ( ! done )
      {
         FD_ZERO(&read_fd);

         FD_SET(fileno(grep_fp), &read_fd);

         tv.tv_sec  = 0;
         tv.tv_usec = 0;

         select(fileno(grep_fp)+1, &read_fd, NULL, NULL, &tv);

         if ( FD_ISSET(fileno(grep_fp), &read_fd) )
         {
            if ( fgets(input, sizeof(input), grep_fp) )
            {
               input[ strlen(input) - 1] = '\0';
               int line_number = 0;

               int offset      = 0;
               int colon_count = 0;
               while ( input[offset] && colon_count < 2)
               {
                  if ( input[offset] == ':' )
                  {
                     if ( colon_count == 0 )
                     {
                        line_number = atoi( &input[offset+1] );
                     }
                     input[offset] = '\0';
                     colon_count++;
                  }
                  offset++;
               }

               while ( isspace( input[offset] ) )
                  offset++;

               if ( colon_count == 2 && line_number != 0 )
                  m_lastItem = new QListViewItem(list, m_lastItem, QString(input), QString::number(line_number), QString( &input[offset] ) );
            }
            else
            {
               done = true;
               delete m_prog;
               m_prog  = NULL;
               grep_fp = NULL;

               cancel_b->setEnabled(false);
               search_b->setEnabled(true);
            }

         }
         else
         {
            done = true;
            read_t->start(500, true);
         }
      }
   }
}

void GrepWin::cancelClicked(void)
{
   read_t->stop();
   if ( m_prog )
   {
      m_prog->killChild();
      delete m_prog;
      m_prog = NULL;
      grep_fp = NULL;
   }

   cancel_b->setEnabled( false );
   search_b->setEnabled( true );
}
