/* mainwin.cpp - Main widget for Bronco
 * 
 *  Bronco File Manager
 * 
 *  Copyright (c) 2000, 2001 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <qinputdialog.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include "about.h"
#include "bronco.h"
#include "dirlist.h"
#include "dirtree.h"
#include "errlist.h"
#include "fifolist.h"
#include "fifotext.h"
#include "filewin.h"
#include "findwin.h"
#include "globwin.h"
#include "license.h"
#include "listedit.h"
#include "prefwin.h"
#include "propwin.h"
#include "tarwin.h"
#include "using.h"

/* For X11 window manager hints */
extern "C" {
#include <X11/Xlib.h>
#include <X11/Xutil.h>
}

ProgressHash MainWin::s_progressHash;
FileHash     MainWin::s_fifoHash;
FifoTextHash MainWin::s_fifoText;
FifoListHash MainWin::s_fifoList;

MainWin::MainWin(QWidget *parent = NULL, char *name = NULL) : QWidget(parent, name)
{
   /* Set up X11 window manager hints, do this before setting the app icon*/
   Display  * display   = x11Display();
   WId        window_id = winId();
   XWMHints * hints     = XGetWMHints( display, window_id ); /* Get hints */
   hints->window_group  = window_id;
   hints->flags         = WindowGroupHint;
   XSetWMHints( display, window_id, hints );  /* set the window hints for WM to use. */
   XFree( hints );

   /* Set application icon */
   setIcon( QPixmap(bronco48_xpm) );
   setIconText( "Bronco File Manager" );

   int x, y;

   QVBoxLayout *vbox = new QVBoxLayout(this);
   QSplitter *splitter = new QSplitter(this);
   splitter->setOrientation(Horizontal);

   setMinimumSize(600, 350);

   tree = new DirTree(splitter, "tree");
   tree->setSorting(-1, FALSE);

   tree->addColumn(NULL);
   tree->setRootIsDecorated(TRUE);

   tree->header()->hide();
   tree->header()->setClickEnabled(FALSE, -1);
   tree->setTreeStepSize(12);

   tree->addRoot("/");

   list = new DirList(splitter, "");
   list->setSorting(-1, FALSE);
   list->loadDir("/");

   glob_win    = new GlobWin(NULL, "");
   pref_win    = new PrefWin(NULL, "");
   license_win = new License(NULL, "");
   about_win   = new AboutWin(NULL, "");
   using_win   = new Using(NULL, "");

   QPopupMenu *launch_menu = list->getLaunchMenu();
   file_menu = new QPopupMenu(this, "file_menu");
   file_menu->insertItem("Open",         list, SLOT(openItem()));
   file_menu->insertItem("Open with...", list, SLOT(openItemWith()));
   file_menu->insertItem("Custom commands",  launch_menu);
   file_menu->insertSeparator();
   file_menu->insertItem("Find...",    list, SLOT(findWindow()),  CTRL+Key_F);
   file_menu->insertItem("Grep...",    list, SLOT(grepWindow()),  CTRL+Key_G);
   file_menu->insertItem("New window", list, SLOT(newWindow()),   CTRL+Key_N);
   file_menu->insertItem("New shell",  list, SLOT(launchShell()), CTRL+Key_S);
   file_menu->insertSeparator();
   file_menu->insertItem("New archive...",   list, SLOT(createTar()), CTRL+Key_T);
   file_menu->insertItem("New directory...", list, SLOT(newDirectory()));
   file_menu->insertItem("New file...",      list, SLOT(newFile()));
   file_menu->insertSeparator();
   file_menu->insertItem("Error logs...",    this, SLOT(showErrorLogs()));
   file_menu->insertSeparator();
   file_menu->insertItem("Quit", qApp, SLOT(quit()), CTRL+Key_Q);

   select_menu = new QPopupMenu(this, "select_menu");
   select_menu->insertItem("Copy",       list, SLOT(copyCommand()),   CTRL+Key_C);
   select_menu->insertItem("Move",       list, SLOT(moveCommand()),   CTRL+Key_X);
   select_menu->insertItem("Paste",      list, SLOT(pasteCommand()),  CTRL+Key_V);
   select_menu->insertItem("Rename...",  list, SLOT(renameCommand()), CTRL+Key_R);
   select_menu->insertItem("Delete",     list, SLOT(deleteCommand()), Key_Delete);
   select_menu->insertSeparator();
   select_menu->insertItem("Copy to...", list, SLOT(copyTo()));
   select_menu->insertItem("Move to...", list, SLOT(moveTo()));
   select_menu->insertSeparator();
   select_menu->insertItem("Properties..",          list, SLOT(properties()));
   select_menu->insertItem("Change permissions...", list, SLOT(chmod()));
   select_menu->insertSeparator();
   select_menu->insertItem("Select all",     list, SLOT(selectAll()), CTRL+Key_A);
   select_menu->insertItem("Unselect all",   list, SLOT(unSelectAll()), CTRL+SHIFT+Key_A);
   select_menu->insertItem("Select mask...", list, SLOT(selectMask()));

   settings_menu = new QPopupMenu(this, "settings_menu");
   sort_menu     = new QPopupMenu(this, "sort_menu");
   name_id  = sort_menu->insertItem("Name",        this, SLOT(sortName()));
   size_id  = sort_menu->insertItem("Size",        this, SLOT(sortSize()));
   perm_id  = sort_menu->insertItem("Permissions", this, SLOT(sortPermissions()));
   mod_id   = sort_menu->insertItem("Modified",    this, SLOT(sortModified()));
   sort_menu->insertSeparator();
   ascii_id = sort_menu->insertItem("Ascii", this, SLOT(sortAscii()));

   settings_menu->insertItem("Sort by", sort_menu);
   settings_menu->insertSeparator();
   settings_menu->insertItem("&File associations...", glob_win, SLOT(openWindow()));
   settings_menu->insertItem("&Preferences...",       pref_win, SLOT(openWindow()));
   settings_menu->setCheckable(TRUE);

   settings_menu->setItemChecked(name_id, TRUE);
   settings_menu->setItemChecked(ascii_id, (prefs.asciiSort()==SORT_ASCII) ? TRUE : FALSE);

   help_menu = new QPopupMenu(this, "help_menu");
   help_menu->insertItem("Using...",   using_win,   SLOT(show()));
   help_menu->insertItem("License...", license_win, SLOT(show()));
   help_menu->insertSeparator();
   help_menu->insertItem("About...",   about_win,   SLOT(show()));

#ifdef USE_KDE
   menu_bar = new KMenuBar(this, "menu_bar");
#else  /* ! USE_KDE */
   menu_bar = new QMenuBar(this, "menu_bar");
#endif /* USE_KDE */
   menu_bar->insertItem("&File", file_menu);
   menu_bar->insertItem("S&election", select_menu);
   menu_bar->insertItem("&Settings", settings_menu);
   menu_bar->insertSeparator();
   menu_bar->insertItem("&Help", help_menu);
   menu_bar->move(0,0);

   vbox->addWidget(menu_bar);
   vbox->addWidget(splitter, 1);

   fifo_t  = new QTimer(this, "");
   child_t = new QTimer(this, "");

   connect(tree,     SIGNAL(currentChanged(DirItem *)), list, SLOT(changeDir(DirItem *)) );
   connect(pref_win, SIGNAL(preferencesChanged()), this, SLOT(refresh()));
   connect(pref_win, SIGNAL(preferencesChanged()), list, SLOT(rebuildLaunchMenu()));
   connect(pref_win, SIGNAL(fontChanged()),        list, SLOT(changeFont()));
   connect(pref_win, SIGNAL(fontChanged()),        this, SLOT(changeFont()));
   connect(fifo_t,   SIGNAL(timeout()), this, SLOT(fifoTimeout()));
   connect(child_t,  SIGNAL(timeout()), this, SLOT(childTimeout()));
   connect(list,     SIGNAL(directoryChanged(const QString &)), this, SLOT(updateCaption(const QString &)));
   connect(list,     SIGNAL(directoryChanged(const QString &)), this, SLOT(dirlistNewDirectory(const QString &)));

   list->setFocus();
   prefs.getBroncoSize(&x, &y);
   resize(x, y);

   QValueList<int> sizes;
   sizes.append(200);
   sizes.append(this->width() - 200);
   splitter->setSizes(sizes);

   fifo_t->start(  100 );  /* 1/10 of a second */
   child_t->start( 1000 ); /* 1 second */

   this->show();
}

void MainWin::resizeEvent(QResizeEvent *e){
   prefs.setBroncoSize(this->width(), this->height());
   prefs.save();
}

void MainWin::sortChange(int id){
   int sort_policy = 0;
   int do_tree = 0;

   if(id == name_id || id == size_id || id == perm_id || id == mod_id){
      settings_menu->setItemChecked(name_id, FALSE);
      settings_menu->setItemChecked(size_id, FALSE);
      settings_menu->setItemChecked(perm_id, FALSE);
      settings_menu->setItemChecked(mod_id,  FALSE);

      settings_menu->setItemChecked(id, TRUE);
   }
   else{
      settings_menu->setItemChecked(id, (settings_menu->isItemChecked(id) ? FALSE : TRUE));
      if(id == ascii_id){
         do_tree = 1;

         if(settings_menu->isItemChecked(ascii_id)){
            prefs.setAsciiSort(true);
         }
         else{
            prefs.setAsciiSort(false);
         }
      }
   }


   if(settings_menu->isItemChecked(name_id)) sort_policy = prefs.asciiSort();
   if(settings_menu->isItemChecked(size_id)) sort_policy = SORT_SIZE;
   if(settings_menu->isItemChecked(perm_id)) sort_policy = SORT_PERM;
   if(settings_menu->isItemChecked(mod_id))  sort_policy = SORT_MTIME;
   if(settings_menu->isItemChecked(rev_id))  sort_policy = (sort_policy | SORT_REVERSE);
   sort_policy = (sort_policy | prefs.getDirListOpts());

   list->sortChange(sort_policy);
   if(do_tree){
      tree->refreshTree();
   }
   return;
}

void MainWin::sortName(void){
   sortChange(name_id);
   return;
}

void MainWin::sortSize(void){
   sortChange(size_id);
   return;
}

void MainWin::sortPermissions(void){
   sortChange(perm_id);
   return;
}

void MainWin::sortModified(void){
   sortChange(mod_id);
   return;
}

void MainWin::sortAscii(void){
   sortChange(ascii_id);
   return;
}

/* Refreshes main window, Directory Tree, and Directory List in 
 * the main window.  Floating directory lists are not affected
 */
void MainWin::refresh(void){
   char *dir;
   dir = strdup(list->getDir());

   int do_tree = 0;
   if(prefs.asciiSort() == SORT_ASCII && settings_menu->isItemChecked(ascii_id))
      do_tree = 1;
   if(prefs.asciiSort() == SORT_ALPHA && !(settings_menu->isItemChecked(ascii_id)))
      do_tree = 1;
   settings_menu->setItemChecked(name_id, TRUE);
   settings_menu->setItemChecked(size_id, FALSE);
   settings_menu->setItemChecked(perm_id, FALSE);
   settings_menu->setItemChecked(mod_id,  FALSE);
   settings_menu->setItemChecked(ascii_id, (prefs.asciiSort()==SORT_ASCII) ? TRUE : FALSE);

   if(do_tree)
   {
      tree->refreshTree();
      list->loadDir(dir);
   }
   else
   {
      list->refresh();
   }

   free(dir);
   
   return;
}

/* Accept command line arguments
 * Treat arguments as directories to open in new windows 
 */
void MainWin::passArgs(int argc, char *argv[]){
   int c;
   int t;
   int mainset = 0;
   const char *offset;
   const char *home = NULL;
   char pwd[1024];
   char buffer[4096];

   home = prefs.homeDir();
   getcwd(pwd, sizeof(pwd) - 1);

   for(c=1; c<argc; c++){
      if(argv[c][0] != '-'){
         if(argv[c][0] == '~'){
            strcpy(buffer, home);
            strcat(buffer, argv[c] + 1);
         }
         else if(argv[c][0] != '/'){
            snprintf(buffer, sizeof(buffer), "%s/", pwd);
            offset = argv[c];
            while(1){
               if((*offset == '.') && (*(offset + 1) == '/')){
                  offset += 2;
               }
               if((*offset == '.') && (*(offset + 1) == '\0')){
                  offset += 1;
               }
               else if((*offset == '.') /* Go up one directory */
                     && (*offset + 1) 
                     && (*(offset + 2) == '/' || *(offset + 2) == '\0')){

                  offset += 2;
                  if(*offset != '\0') offset++;

                  t = strlen(buffer) - 1;
                  if(t > 0) buffer[t] = '\0';
                  while(t > 0){
                     if(buffer[t] == '/'){
                        break;
                     }
                     else{
                        buffer[t] = '\0';
                     }
                     t--;
                  }
               }
               else{
                  break;
               }
            }
            strcat(buffer, offset);
         }
         else{
            strcpy(buffer, argv[c]);
         }

         if(mainset){
            list->newWindow(buffer);
         }
         else{
            tree->setDir(buffer);
            mainset = 1;
         }
      }
   }

   if(!mainset){
      char *dir;
      if((dir = getcwd(NULL, 0)) != NULL){
         tree->setDir(dir);
         free(dir);
      }
   }
}

/* Check for messages from children or helper apps
 * and respond to them appropriately
 *
 * I should probably break this up into sub functions... color me lazy
 */
void MainWin::fifoTimeout()
{
   char ch;
   char input[4096];

   bool done = false;
   while ( !done )
   {
      done = true;

      input[0] = '\0';
      for( int t=0; read(fifo_fd, &ch, 1) > 0; t++ )
      {
         input[t] = ch;
         if ( ch =='\r' || ch == '\n')
         {
            input[t] = '\0';
            break;
         }

         if ( t == sizeof(input) - 1)
         {
            input[t] = '\0';
            break;
         }
      }

      if ( input[0] )
      {
         if ( helper_debug )
            printf("%s\n", input);

         switch ( input[0] )
         {
            case '0':

               if( input[1] == '0' )
               {
                  if ( helper_debug )
                     printf("  Open child fifo\n");

                  int child_pid = atoi( &input[2] );
                  QString filename;
                  filename.sprintf("%s/.bronco/child%d", prefs.homeDir(), child_pid );
                  
                  /* Our helper has a limited amount of time to open 
                   * the FIFO. */
                  int fd = 0;
                  int count = 0;
                  while ( fd <= 0)
                  {
                     fd = open(filename.latin1(), O_WRONLY | O_NONBLOCK);

                     if ( fd <= 0 )
                     {
                        count++;
                        if ( count > 1 )
                        {
                           if ( helper_debug )
                              printf("No reader for fifo %s\n", filename.latin1());
                           break;
                        }
                        sleep(1);
                     }
                  }

                  if ( fd > 0 )
                  {
                     FILE *fp = fdopen(fd, "w");

                     if ( fp != NULL )
                     {
                        s_fifoHash[child_pid] = fp;
                        if ( helper_debug )
                           printf("Fifo opened for %d\n", child_pid);
                     }
                     else
                     {
                        if ( helper_debug )
                           printf("Fifo for %d: %s\n", child_pid, strerror(errno));
                     }
                  }
                  break;
               }
               else
               {
                  if ( helper_debug)
                     printf("Unknown command: %c%c\n", input[0], input[1]);
               }
               break;


            case '1':  /* Messages for user */

               if( input[1] == '0' )
               {
                  if ( helper_debug )
                     printf("  Show Information Dialog\n");

                  int child_pid = 0;
                  char message[80];
                  sscanf(&input[3], "%d %80[^\n]s", &child_pid, message);

                  QMessageBox::information(NULL, "Bronco", message);
               }
               else if ( input[1] == '1' )
               {
                  if ( helper_debug )
                     printf("  Show Warning Dialog\n");

                  int child_pid = 0;
                  char message[80];
                  sscanf(&input[3], "%d %80[^\n]s", &child_pid, message);

                  QMessageBox::warning(NULL, "Bronco", message);
               }
               else if ( input[1] == '2' )
               {
                  if ( helper_debug )
                     printf("  Show Critical Error Dialog\n");

                  int child_pid = 0;
                  char message[80];
                  sscanf(&input[3], "%d %80[^\n]s", &child_pid, message);

                  QMessageBox::critical(NULL, "Bronco", message);
               }
               else
               {
                  if ( helper_debug)
                     printf("Unknown command: %c%c\n", input[0], input[1]);
               }
               break;


            case '2':  /* Questions for user */

               if( input[1] == '0' )
               {
                  if ( helper_debug )
                     printf("  Yes or No Question\n");

                  int child_pid = 0;
                  char message[80];
                  sscanf(&input[3], "%d %80[^\n]s", &child_pid, message);

                  if ( s_fifoHash[child_pid] )
                  {
                     FILE *fp = s_fifoHash[child_pid];

                     int r = QMessageBox::information(NULL, "Bronco", message, QMessageBox::Yes | QMessageBox::Default, QMessageBox::No | QMessageBox::Escape, 0);

                     if ( r == QMessageBox::Yes)
                     {
                        fprintf(fp, "yes\n");
                        fflush(fp);
                     }
                     else if ( r == QMessageBox::No)
                     {
                        fprintf(fp, "no\n");
                        fflush(fp);
                     }
                  }
                  else
                  {
                     if ( helper_debug )
                        printf("No fifo for %d\n", child_pid);
                  }
               }
               else if ( input[1] == '1' )
               {
                  if ( helper_debug )
                     printf("  Yes, No, or Cancel Question\n");

                  int child_pid = 0;
                  char message[80];
                  sscanf(&input[3], "%d %80[^\n]s", &child_pid, message);

                  if ( s_fifoHash[child_pid] )
                  {
                     FILE *fp = s_fifoHash[child_pid];

                     int r = QMessageBox::information(NULL, "Bronco", message, QMessageBox::Yes | QMessageBox::Default, QMessageBox::No, QMessageBox::Cancel | QMessageBox::Escape);

                     if ( r == QMessageBox::Yes)
                     {
                        fprintf(fp, "yes\n");
                        fflush(fp);
                     }
                     else if ( r == QMessageBox::No)
                     {
                        fprintf(fp, "no\n");
                        fflush(fp);
                     }
                     else if ( r == QMessageBox::Cancel)
                     {
                        fprintf(fp, "cancel\n");
                        fflush(fp);
                     }
                  }
                  else
                  {
                     if ( helper_debug )
                        printf("No fifo for %d\n", child_pid);
                  }
               }
               else if ( input[1] == '5' )
               {
                  if ( helper_debug )
                     printf("  Get Text From User\n");

                  int child_pid = 0;
                  char message[80];
                  sscanf(&input[3], "%d %80[^\n]s", &child_pid, message);

                  if ( s_fifoHash[child_pid] )
                  {
                     FILE *fp = s_fifoHash[child_pid];
                     bool ok;

//                     QString answer = QInputDialog::getText("Bronco", message, QString::null, &ok);
                     QString answer = QInputDialog::getText( "Bronco", message, QLineEdit::Normal, QString::null, &ok);

                     if ( !answer.isNull() && !answer.isEmpty() )
                     {
                        fprintf(fp, "%s\n", answer.latin1() );
                     }
                     else
                     {
                        fprintf(fp, "\n");
                     }
                     fflush(fp);
                  }
                  else
                  {
                     if ( helper_debug )
                        printf("No fifo for %d\n", child_pid);
                  }
               }
               else if ( input[1] == '6' )
               {
                  if ( helper_debug )
                     printf("  Get Integer From User\n");

                  int child_pid = 0;
                  char message[80];
                  sscanf(&input[3], "%d %80[^\n]s", &child_pid, message);

                  if ( s_fifoHash[child_pid] )
                  {
                     FILE *fp = s_fifoHash[child_pid];

                     int answer = QInputDialog::getInteger("Bronco", message);

                     fprintf(fp, "%d\n", answer );
                     fflush(fp);
                  }
                  else
                  {
                     if ( helper_debug )
                        printf("No fifo for %d\n", child_pid);
                  }
               }
               else if ( input[1] == '7' )
               {
                  if ( helper_debug )
                     printf("  Get Floating-Point Number From User\n");

                  int child_pid = 0;
                  char message[80];
                  sscanf(&input[3], "%d %80[^\n]s", &child_pid, message);

                  if ( s_fifoHash[child_pid] )
                  {
                     FILE *fp = s_fifoHash[child_pid];

                     double answer = QInputDialog::getDouble("Bronco", message);
                     
                     fprintf(fp, "%f\n", (float) answer );
                     fflush(fp);
                  }
                  else
                  {
                     if ( helper_debug )
                        printf("No child for %d\n", child_pid);
                  }
               }
               else if ( input[1] == '8' )
               {
                  if ( helper_debug )
                     printf("  Get Combo Box Selection From User\n");

                  int child_pid = 0;
                  char message[80];
                  child_pid = atoi( &input[3] );

                  QString question = "";
                  QStringList answers;
                  
                  int n = 3;

                  while ( input[n] && isspace( input[n]) )
                     n++;
                  while ( input[n] && !isspace( input[n]) )
                     n++;
                  while ( input[n] && isspace( input[n]) )
                     n++;

                  while ( input[n] )
                  {
                     sscanf(&input[n], "%80[^\n|]s", message);

                     if ( question.isEmpty() )
                     {
                        question = QString(message);
                     }
                     else
                     {
                        answers.append( QString(message) );
                     }

                     while ( input[n] != '\0' && input[n] != '|' )
                        n++;

                     if ( input[n] )
                        n++;
                  }

                  if ( s_fifoHash[child_pid] )
                  {
                     FILE *fp = s_fifoHash[child_pid];
                     bool ok = false;

                     QString answer = QInputDialog::getItem("Bronco", question, answers, 0, false, &ok);
                     
                     if ( ok )
                     {
                        fprintf(fp, "%s\n", answer.latin1() );
                     }
                     else
                     {
                        fprintf(fp, "\n");
                     }
                     fflush(fp);
                  }
                  else
                  {
                     if ( helper_debug )
                        printf("No child for %d\n", child_pid);
                  }
               }
               else
               {
                  if ( helper_debug)
                     printf("Unknown command: %c%c\n", input[0], input[1]);
               }
               break;


            case '3':

               if ( input[1] == '0' )
               {
                  if ( helper_debug )
                     printf("  Create Progress Dialog\n");

                  char name[80];
                  int win_num = 0;
                  name[0] = '\0';

                  sscanf(&input[2], "%d %80[^\n]s", &win_num, name);
                  if ( s_progressHash[win_num] )
                  {
                     delete s_progressHash[win_num];
                     s_progressHash.erase(win_num);
                  }

                  if ( helper_debug )
                     printf("Creating progress dialog for %d\n", win_num);

                  s_progressHash[win_num] = new QProgressDialog(NULL, "");
                  s_progressHash[win_num]->setAutoReset( false );

                  if ( name[0] )
                  {
                     s_progressHash[win_num]->setCaption( QString(name) );
                  }
                  else
                  {
                     s_progressHash[win_num]->setCaption( "Bronco Helper" );
                  }
               }
               else if ( input[1] == '1' )
               {
                  if ( helper_debug )
                     printf("  Set Progress Percentage\n");

                  int win_num = 0, p = 0;
                  sscanf(&input[2], "%d %d", &win_num, &p);

                  if ( p > 100 )
                    p = 100;

                  if ( helper_debug )
                     printf("Setting progress to %d for %d\n", p, win_num);

                  if ( s_progressHash[win_num] && (p > 0 && p <= 100) )
                  {
                     s_progressHash[win_num]->setProgress(p);
                  }
               }
               else if ( input[1] == '2' )
               {
                  if ( helper_debug )
                     printf("  Set Progress Label Text\n");

                  char label[80];
                  int win_num = 0;
                  sscanf(&input[2], "%d %80[^\n]s", &win_num, label);

                  if ( helper_debug )
                     printf("Setting label to '%s' for %d\n", label, win_num);

                  if ( s_progressHash[win_num] )
                  {
                     s_progressHash[win_num]->setLabelText( QString(label) );
                  }
               }
               else if ( input[1] == '3' )
               {
                  if ( helper_debug )
                     printf("  Close Progress Dialog Box\n");

                  int win_num = 0;
                  sscanf(&input[2], "%d", &win_num);

                  if ( helper_debug )
                     printf("Closing progress dialog for %d\n", win_num);

                  if ( s_progressHash[win_num] )
                  {
                     delete s_progressHash[win_num];
                     s_progressHash.erase(win_num);
                  }
               }
               else
               {
                  if ( helper_debug)
                     printf("Unknown command: %c%c\n", input[0], input[1]);
               }
               break;


            case '4':

               if ( input[1] == '0' )
               {
                  if ( helper_debug )
                     printf("  Create Text Window\n");

                  char name[80];
                  int win_num = 0;
                  name[0] = '\0';

                  sscanf(&input[2], "%d %80[^\n]", &win_num, name);

                  if ( helper_debug )
                     printf("Creating text window for %d\n", win_num);

                  if ( s_fifoText[win_num] )
                  {
                     FifoText *doomed = s_fifoText[win_num];
                     s_fifoText.erase(win_num);
                     delete doomed;
                  }
                  s_fifoText[win_num] = new FifoText( win_num );
                  connect( s_fifoText[win_num], SIGNAL(closeFifoText(int)),
                        this, SLOT(closeFifoText(int)));
                  if ( name[0] )
                  {
                     s_fifoText[win_num]->setCaption( QString(name) );
                  }
                  else
                  {
                     s_fifoText[win_num]->setCaption( "Bronco Helper" );
                  }
               }
               else if ( input[1] == '1' )
               {
                  if ( helper_debug )
                     printf("  Append Text\n");

                  char name[80];
                  int win_num = 0;
                  name[0] = '\0';

                  sscanf(&input[2], "%d %80[^\n]", &win_num, name);

                  if ( helper_debug )
                     printf("Appending text for %d\n", win_num);

                  if ( s_fifoText[win_num] )
                  {
                     s_fifoText[win_num]->appendText(name);
                  }
                  else
                  {
                     if ( helper_debug )
                        printf("No text window for %d\n", win_num);
                  }
               }
               else if ( input[1] == '5' )
               {
                  if ( helper_debug )
                     printf("  Create List Window\n");

                  char name[80];
                  int win_num = 0;
                  name[0] = '\0';

                  sscanf(&input[2], "%d %80[^\n]", &win_num, name);

                  if ( helper_debug )
                     printf("Creating list window for %d\n", win_num);

                  if ( s_fifoList[win_num] )
                  {
                     FifoList *doomed = s_fifoList[win_num];
                     s_fifoList.erase(win_num);
                     delete doomed;
                  }
                  s_fifoList[win_num] = new FifoList( win_num );
                  connect( s_fifoList[win_num], SIGNAL(closeFifoList(int)),
                        this, SLOT(closeFifoList(int)));
                  if ( name[0] )
                  {
                     s_fifoList[win_num]->setCaption( QString(name) );
                  }
                  else
                  {
                     s_fifoList[win_num]->setCaption( "Bronco Helper" );
                  }
               }
               else if ( input[1] == '6' )
               {
                  if ( helper_debug )
                     printf("  Add Column\n");

                  char name[80];
                  int win_num = 0;
                  name[0] = '\0';

                  sscanf(&input[2], "%d %80[^\n]", &win_num, name);

                  if ( helper_debug )
                     printf("Adding list column for %d\n", win_num);

                  if ( s_fifoList[win_num] )
                  {
                     s_fifoList[win_num]->addColumn(name);
                  }
                  else
                  {
                     if ( helper_debug )
                        printf("No window list for %d\n", win_num);
                  }
               }
               else if ( input[1] == '7' )
               {
                  if ( helper_debug )
                     printf("  Append List Items\n");

                  int child_pid = 0;
                  char message[4096];
                  child_pid = atoi( &input[3] );

                  QStringList elements;
                  
                  int n = 3;

                  while ( input[n] && isspace( input[n]) )
                     n++;
                  while ( input[n] && !isspace( input[n]) )
                     n++;
                  while ( input[n] && isspace( input[n]) )
                     n++;

                  while ( input[n] )
                  {
                     sscanf(&input[n], "%4096[^\n|]s", message);

                     if ( helper_debug )
                        printf("item: %s\n", message);
                     elements.append( QString(message) );

                     while ( input[n] != '\0' && input[n] != '|' )
                        n++;

                     if ( input[n] )
                        n++;
                  }

                  if ( helper_debug )
                     printf("Appending items for %d\n", child_pid);

                  if ( s_fifoList[child_pid] )
                  {
                     s_fifoList[child_pid]->appendItem( elements );
                  }
                  else
                  {
                     if ( helper_debug )
                        printf("No window list for %d\n", child_pid);
                  }
               }
               else
               {
                  if ( helper_debug)
                     printf("Unknown command: %c%c\n", input[0], input[1]);
               }
               break;


            case '5': /* Directory commands */

               if( input[1] == '0' )
               {
                  if ( helper_debug )
                     printf("  Open New Directory List Window\n");

                  char dir[4096];
                  int child_pid = 0;
                  dir[0] = '\0';

                  sscanf(&input[2], "%d %4096[^\n]s", &child_pid, dir);
                  if ( dir[0] )
                  {
                     list->newWindow( dir );
                  }

                  if ( helper_debug )
                     printf("Opening %s for %d\n", dir, child_pid);
               }
               else if( input[1] == '1' )
               {
                  if ( helper_debug )
                     printf("  Refresh Directory List Windows\n");

                  char dir[4096];
                  int child_pid = 0;
                  dir[0] = '\0';

                  sscanf(&input[2], "%d %4096[^\n]s", &child_pid, dir);
                  emit refreshDirectory( QString(dir) );

                  if ( helper_debug )
                     printf("Refreshing %s for %d\n", dir, child_pid);
               }
               else
               {
                  if ( helper_debug)
                     printf("Unknown command: %c%c\n", input[0], input[1]);
               }
               break;


            case '8': /* File progress (internal) */

               if ( helper_debug)
                  printf("  Internal command\n");

               if ( input[1] == '0' )
               {
                  int code, win_num, progress;
                  sscanf(input, "%d %d %d", &code, &win_num, &progress);
                  FileWin *win;
                  win = FileWin::getWinNumber( win_num );
                  if ( win )
                  {
                     win->setProgress(progress);
                  }
               }
               else if ( input[1] == '1' )
               {
                  int code, win_num;
                  sscanf(input, "%d %d", &code, &win_num);
                  FileWin *win;
                  win = FileWin::getWinNumber( win_num );
                  if ( win )
                  {
                     win->goAway();
//                     refresh();
                  }
               }
               else if ( input[1] == '2' )
               {
                  char filename[4096];
                  int code = 0, win_num = 0;
                  sscanf(input, "%d %d %4096[^\n]s", &code, &win_num, filename);

                  FileWin *win;
                  win = FileWin::getWinNumber( win_num );
                  if ( win )
                  {
                     win->setCurrentFile(filename);
                  }
               }
               else
               {
                  if ( helper_debug)
                     printf("Unsupported command: %c%c\n", input[0], input[1]);
               }
               break;


            default:

               if ( helper_debug)
                  printf("Unsupported command: %c%c\n", input[0], input[1]);
               break;

         }
         done = false;
      }
   }
}

void MainWin::updateCaption(const QString &s)
{
   QString caption = s;
   caption.prepend("Bronco - ");
   setCaption(caption);
}

void MainWin::dirlistNewDirectory(const QString &dir)
{
   tree->setDir(dir.latin1());
}

/* Delete list window when user is done */
void MainWin::closeFifoList( int child_pid )
{
   if ( s_fifoList[child_pid] )
   {
      FifoList *doomed = s_fifoList[child_pid];
      s_fifoList.erase(child_pid);
      delete doomed;
   }
}

/* Delete text window when user is done */
void MainWin::closeFifoText( int child_pid )
{
   if ( s_fifoText[child_pid] )
   {
      FifoText *doomed = s_fifoText[child_pid];
      s_fifoText.erase(child_pid);
      delete doomed;
   }
}

/* Reap child processes, and clean up after the spoiled brats */
void MainWin::childTimeout(void)
{
   int status    = 0;
   int child_pid = 0;

   while ( (child_pid = waitpid(-1, &status, WNOHANG)) > 0 )
   {
      childCleanup( child_pid );
   }
}

void MainWin::childCleanup(int child_pid)
{
   QString file;
   const char *home;

   home = prefs.homeDir();

   file = QString(home) + QString("/.bronco/child") + QString::number(child_pid);

   unlink(file.latin1()); /* remove stale fifo if it exists */

   /* delete progress dialog if it exists */
   if ( s_progressHash[child_pid] )
   {
      QProgressDialog* doomed = s_progressHash[child_pid];
      s_progressHash.erase(child_pid);
      delete doomed;
   }

   file = QString(home) + QString("/.bronco/errorlog") + QString::number(child_pid);

   /* View error log if it exists */

   if ( prefs.getShowErrors() )
   {
      FILE *fp;

      if ( (fp = fopen(file.latin1(), "r")) != NULL )
      {
         char input[1024];

         if ( s_fifoText[child_pid] )
         {
            FifoText * doomed = s_fifoText[child_pid];
            s_fifoText.erase(child_pid);
            delete doomed;
         }

         FifoText *text = new FifoText(child_pid);
         s_fifoText[child_pid] = text;
         connect( s_fifoText[child_pid], SIGNAL(closeFifoText(int)),
               this, SLOT(closeFifoText(int)));

         text->setCaption(QString("Error Log for ") + QString::number(child_pid));
         while ( fgets(input, sizeof(input), fp) )
         {
            int n = strlen(input) - 1;
            while ( n >= 0 && input[n] == '\n')
               input[n] = '\0';
            text->appendText(input);
         }
         fclose(fp);
      }
   }

   /* Delete error log if it exists */
//   unlink( file.latin1() );
}

PrefWin * MainWin::getPrefWin(void)
{
   return pref_win;
}

void MainWin::showErrorLogs(void)
{
   ErrList *err = new ErrList(NULL, "", WDestructiveClose);
   err->show(); /* I do this to avoid a "not used" warning */
}

void MainWin::changeFont(void)
{
   char family[64];
   int  fsize = 10;

   prefs.getFont(family, &fsize);
   QFont font(family, fsize);
   tree->setFont( font );
}
